/*
 * LMDB Abstraction Layer.
 * Copyright (C) 2023  Yury Gubich <blue@macaw.me>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef LMDBAL_CACHE_H
#define LMDBAL_CACHE_H

#include <map>
#include <set>
#include <tuple>

#include "storage.h"

namespace LMDBAL {

template <class K, class V>
class Cache : public Storage<K, V> {
    friend class Base;
    enum class Mode {       //it's a cache state when we:
        nothing,            //  - know nothing about records in database on disk
        size,               //  - know just an amount of records
        full                //  - shure that our cache is equal to the database on disk
    };

    enum class Operation {
        add,
        remove,
        change,
        force,
        drop,
        replace,
        addMany
    };

    typedef std::pair<Operation, void*> Entry;
    typedef std::list<Entry> Queue;
    typedef std::map<TransactionID, Queue> TransactionCache;

protected:
    Cache(const std::string& name, Base* parent);
    ~Cache() override;

    virtual void transactionStarted(TransactionID txn, bool readOnly) const override;
    virtual void transactionCommited(TransactionID txn) override;
    virtual void transactionAborted(TransactionID txn) const override;
private:
    void handleMode() const;

    void handleTransactionEntry(const Entry& entry);
    void destroyTransactionEntry(const Entry& entry) const;

    void handleAddRecord(const K& key, const V& value);
    void handleRemoveRecord(const K& key);
    void handleChangeRecord(const K& key, const V& value);
    void handleForceRecord(const K& key, const V& value, bool added);
    void handleReplaceAll(std::map<K, V>* data);
    void handleAddRecords(const std::map<K, V>& data, bool overwrite, SizeType newSize);
    void handleDrop();

public:
    using Storage<K, V>::drop;
    virtual int drop(TransactionID transaction) override;
    virtual void addRecord(const K& key, const V& value) override;
    virtual void addRecord(const K& key, const V& value, TransactionID txn) override;
    virtual bool forceRecord(const K& key, const V& value) override;
    virtual bool forceRecord(const K& key, const V& value, TransactionID txn) override;
    virtual void changeRecord(const K& key, const V& value) override;
    virtual void changeRecord(const K& key, const V& value, TransactionID txn) override;
    virtual void removeRecord(const K& key) override;
    virtual void removeRecord(const K& key, TransactionID txn) override;
    virtual bool checkRecord(const K& key) const override;
    virtual bool checkRecord(const K& key, TransactionID txn) const override;
    virtual V getRecord(const K& key) const override;
    virtual V getRecord(const K& key, TransactionID txn) const override;
    virtual SizeType count() const override;
    virtual SizeType count(TransactionID txn) const override;

    virtual std::map<K, V> readAll() const override;
    virtual void replaceAll(const std::map<K, V>& data) override;
    virtual SizeType addRecords(const std::map<K, V>& data, bool overwrite = false) override;

protected:
    Mode* mode;
    std::map<K, V>* cache;
    std::set<K>* abscent;
    SizeType* sizeDifference;
    TransactionCache* transactionCache;
};

}

#include "cache.hpp"

#endif // LMDBAL_CACHE_H
